@@ -6,7 +6,8 @@ import VITAE.model as model
 import VITAE.preprocess as preprocess
 import VITAE.train as train
 from VITAE.inference import Inferer
-from VITAE.utils import load_data, get_embedding, get_igraph, louvain_igraph, plot_clusters, plot_marker_gene
+from VITAE.utils import load_data, get_embedding, get_igraph, louvain_igraph, \
+    plot_clusters, plot_marker_gene, DE_test
 from VITAE.metric import topology, get_GRI
 
 from sklearn.metrics.cluster import adjusted_rand_score
@@ -16,6 +17,7 @@ import numpy as np
 import scipy as sp
 import pandas as pd
 import networkx as nx
+from scipy import stats
 
 
 class VITAE():
@@ -27,7 +29,17 @@ class VITAE():
 
     def get_data(self, X = None, adata = None, labels = None,
                  covariate = None, cell_names = None, gene_names = None):
-        '''Get data for model.
+        '''Get data for model. 
+        
+        (1) The user can provide a 2-dim numpy array as the count matrix `X`, either preprocessed or raw. 
+        Some extra information `labels`, `cell_names` and `gene_names` (as 1-dim numpy arrays) are optional.
+
+        (2) If the package `scanpy` is installed, then the function can also accept an `annData` input as `adata`. 
+        Some extra information `labels`, `cell_names` and `gene_names` are extracted from 
+        `adata.obs.cell_types`, `adata.obs_names.values` and `adata.var_names.values`, and
+        a 1-dim numpy array `labels` can also be provided if `adata.obs.cell_types` does not exist.
+
+        Covariates can be provided as a 2-dim numpy array.
 
         Parameters
         ----------
@@ -67,12 +79,16 @@ class VITAE():
         
     def preprocess_data(self, processed: bool = False, dimred: bool = False,
                         K: float = 1e4, gene_num: int = 2000, data_type: str = 'UMI', npc: int = 64):
-        ''' Data preprocessing - log-normalization, feature selection, and scaling.            
+        ''' Data preprocessing - log-normalization, feature selection, and scaling.                    
+
+        If the inputs are preprocessed by users, then `Gaussian` model will be used and PCA will be performed to reduce the input dimension.
+        Otherwise, preprocessing will be performed on `X` following Seurat's routine. 
+        If `adata` is provided, the preprocession will be done via `scanpy`.
 
         Parameters
         ----------
         processed : boolean, optional
-            Whether adata has been processed.
+            Whether adata has been processed. If `processed=True`, then `Gaussian` model will be used.
         dimred : boolean, optional
             Whether the processed adata is after dimension reduction.
         K : float, optional              
@@ -151,16 +167,14 @@ class VITAE():
             with open(path_to_file+'.label', 'wb') as f:
                 np.save(f, self.cluster_labels)
         with open(path_to_file+'.config', 'wb') as f:
-            np.save(f, [self.dim_origin,
-                        self.dimensions,
-                        self.dim_latent,
-                        self.data_type,
-                        False if self.c_score is None else True])
+            np.save(f, np.array([
+                self.dim_origin, self.dimensions, self.dim_latent,
+                self.data_type, False if self.c_score is None else True], dtype=object))
         if hasattr(self, 'pi'):
             with open(path_to_file+'.inference', 'wb') as f:
-                np.save(f, [self.pi, self.mu, self.pc_x,
-                            self.w_tilde, self.var_w_tilde,
-                            self.D_JS, self.z, self.embed_z, self.inferer.embed_mu])
+                np.save(f, np.array([
+                    self.pi, self.mu, self.pc_x, self.w_tilde, self.var_w_tilde,
+                    self.D_JS, self.z, self.embed_z, self.inferer.embed_mu], dtype=object))
     
 
     def load_model(self, path_to_file: str = 'model.checkpoint', load_labels: bool = False):
@@ -240,7 +254,6 @@ class VITAE():
         path_to_weights : str, optional 
             The path of weight file to be saved; not saving weight if None.
         '''                    
-        train.clear_session()
         if stratify is None:
             stratify = self.labels
         elif stratify is False:
@@ -578,9 +591,41 @@ class VITAE():
                                                          cutoff, 
                                                          path=path, 
                                                          is_plot=is_plot)
+        self.pseudotime = pseudotime
         return G, w, pseudotime
 
-    
+
+    def differentially_expressed_test(self, alpha: float = 0.05):
+        '''Differentially gene expression test. All (selected and unselected) genes will be tested 
+        Only cells in `selected_cell_subset` will be used, which is useful when one need to
+        test differentially expressed genes on a branch of the inferred trajectory.
+
+        Parameters
+        ----------
+        alpha : float, optional
+            The cutoff of p-values.
+
+        Returns
+        ----------
+        res_df : pandas.DataFrame
+            The test results of expressed genes with two columns,
+            the estimated coefficients and the adjusted p-values.
+        '''
+        if not hasattr(self, 'pseudotime'):
+            raise ReferenceError("Pseudotime does not exist! Please run 'infer_trajectory' first.")
+
+        # Prepare X and Y for regression expression ~ rank(PDT) + covariates
+        Y = self.expression[self.selected_cell_subset_id,:]
+        std_Y = np.std(Y, ddof=1, axis=0, keepdims=True)
+        Y = np.divide(Y-np.mean(Y, axis=0, keepdims=True), std_Y, out=np.empty_like(Y)*np.nan, where=std_Y!=0)
+        X = stats.rankdata(self.pseudotime[self.selected_cell_subset_id])
+        X = ((X-np.mean(X))/np.std(X, ddof=1)).reshape((-1,1))
+        X = np.c_[np.ones_like(X), X, self.c_score[self.selected_cell_subset_id,:]]
+
+        res_df = DE_test(Y, X, self.gene_names, alpha)
+        return res_df
+
+
     def plot_marker_gene(self, gene_name: str, refit_dimred: bool = False, dimred: str = 'umap', path: Optional[str] =None, **kwargs):
         '''Plot expression of the given marker gene.
 
